# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ...helpers.validators import ArgumentType
from ...settings import conf

from .base import Command


class NtpMeasureCommand(Command):
    DESCRIPTION = "Create an NTP measurement and wait for the results"

    def add_arguments(self):

        Command.add_arguments(self)

        self.add_primary_argument(name="target", parser=self.parser)

        spec = conf["specification"]["types"]["ntp"]

        specific = self.parser.add_argument_group("NTP-specific Options")
        specific.add_argument(
            "--packets",
            type=ArgumentType.integer_range(minimum=1, maximum=16),
            default=spec["packets"],
            help="The number of packets sent",
        )
        specific.add_argument(
            "--timeout",
            type=ArgumentType.integer_range(minimum=1, maximum=60000),
            default=spec["timeout"],
            help="The timeout per-packet",
        )

    def _get_measurement_kwargs(self):

        r = Command._get_measurement_kwargs(self)

        r["packets"] = self.arguments.packets
        r["timeout"] = self.arguments.timeout

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import itertools


class ValueKeyAggregator(object):
    """Aggregator based on tha actual value of the key/attribute"""

    def __init__(self, key, prefix=None):
        self.aggregation_keys = key.split(".")
        self.key_prefix = prefix or self.aggregation_keys[-1].upper()

    def get_key_value(self, entity):
        """
        Returns the value of the key/attribute the aggregation will use to
        bucketize probes/results
        """
        attribute = entity
        for key in self.aggregation_keys:
            attribute = getattr(attribute, key)
        return attribute

    def get_bucket(self, entity):
        """
        Returns the bucket the specific entity belongs to based on the give
        key/attribute
        """
        return "{0}: {1}".format(self.key_prefix, self.get_key_value(entity))


class RangeKeyAggregator(ValueKeyAggregator):
    """
    Aggregator based on where the position of the value of the key/attribute is
    in the given range
    """

    def __init__(self, key, ranges):
        ValueKeyAggregator.__init__(self, key)
        self.aggregation_ranges = sorted(ranges, reverse=True)

    def get_bucket(self, entity):
        """
        Returns the bucket the specific entity belongs to based on the give
        key/attribute
        """

        bucket = "{0}: < {1}".format(
            self.key_prefix, self.aggregation_ranges[-1]
        )

        key_value = self.get_key_value(entity)
        for index, krange in enumerate(self.aggregation_ranges):
            if key_value > krange:
                if index == 0:
                    bucket = "{0}: > {1}".format(self.key_prefix, krange)
                else:
                    bucket = "{0}: {1}-{2}".format(
                        self.key_prefix,
                        krange,
                        self.aggregation_ranges[index - 1],
                    )
                break

        return bucket


def _get_sort_key(kv):
    key = []
    for is_digit, part in itertools.groupby(kv[0], key=str.isdigit):
        part = "".join(part)
        if is_digit:
            part = int(part)
        key.append(part)
    return key


def aggregate(entities, aggregators):
    """
    Aggregate the given entities using the given aggregators.

    Returns a dict of {combined_aggregation_key_tuple: entity_list}, where
    the keys are in ascending numeric >> lexical order.
    """
    if not aggregators:
        return entities

    buckets = {}

    for e in entities:
        key = " | ".join(a.get_bucket(e) for a in aggregators)
        bucket = buckets.setdefault(key, [])
        bucket.append(e)

    return dict(sorted(buckets.items(), key=_get_sort_key))



# Copyright (c) 2023 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import csv
import re
from io import StringIO
from typing import (Any, Dict, Iterable, Iterator, List, Mapping, Optional,
                    Tuple)

from typing_extensions import Literal, NotRequired, TypedDict

from .colours import colourise

Alignment = Literal["<", "^", ">"]
LEFT: Alignment = "<"
CENTRE: Alignment = "^"
RIGHT: Alignment = ">"


class ColumnDef(TypedDict):
    align: Alignment
    width: int


class RowDef(TypedDict):
    obj: NotRequired[Any]
    values: Dict[str, Any]
    colour: NotRequired[str]


# ((1, "Value1"), (3, True), (4, 5342),)
AggregationKey = Tuple[Any, ...]


class SortableNull:
    def __str__(self):
        return "null"

    def __hash__(self):
        return 0

    def __eq__(self, other):
        return isinstance(other, type(self))

    def __lt__(self, other):
        if isinstance(other, type(self)):
            return False
        return True

    def __gt__(self, other):
        return False


class Renderer:
    """
    Abstract base class for tabular renderers.
    """

    def __init__(
        self,
        rows: List[RowDef],
        total_count: int,
        columns: Dict[str, ColumnDef],
        filters: Mapping[str, str],
        arguments: argparse.Namespace,
    ):
        self.rows = rows
        self.total_count = total_count
        self.columns = columns
        self.filters = filters

        self.aggregate_by: List[str] = arguments.aggregate_by
        self.max_per_aggregation: Optional[int] = arguments.max_per_aggregation
        self.no_header: bool = arguments.no_header

    def __iter__(self) -> Iterator[str]:
        if not self.no_header:
            for line in self.get_header():
                yield line
        if self.aggregate_by:
            for composite, rows in self._aggregate():
                aggr_header = self.get_aggregation_header(composite)
                if aggr_header is not None:
                    yield aggr_header
                for row in rows:
                    yield self.get_line(row)
        else:
            for row in self.rows:
                yield self.get_line(row)
        if not self.no_header:
            for line in self.get_footer():
                yield line

    def get_header(self) -> Iterable[str]:
        """
        Return an iterable of lines to output before the table.
        """
        return []

    def get_aggregation_header(self, composite: AggregationKey) -> Optional[str]:
        """
        Return a line to output before each aggregation bucket.
        """
        return None

    def get_line(self, row: RowDef) -> str:
        """
        Return a single line rendering the given row.
        """
        raise NotImplementedError

    def get_footer(self) -> Iterable[str]:
        """
        Return an iterable of lines to output after the table.
        """
        return []

    def _aggregate(self) -> Iterable[Tuple[AggregationKey, List[RowDef]]]:
        buckets: Dict[AggregationKey, List[RowDef]] = {}
        for row in self.rows:
            values = [row["values"][k] for k in self.aggregate_by]
            composite = tuple(SortableNull() if v is None else v for v in values)
            bucket = buckets.setdefault(composite, [])
            if (
                self.max_per_aggregation is None
                or len(bucket) < self.max_per_aggregation
            ):
                bucket.append(row)
        return sorted(buckets.items())


class PrettyRenderer(Renderer):
    """
    Renderer which outputs colourised, human-readable tables
    """

    _fmt = None

    @property
    def fmt(self) -> str:
        if not self._fmt:
            self._fmt = " ".join(
                f"{{!s:{c['align']}{c['width']}}}" for c in self.columns.values()
            )
        return self._fmt

    def get_header(self) -> Iterable[str]:
        if self.filters:
            yield ""
            yield colourise("Filters:", "white")
            for k, v in self.filters.items():
                if k not in ("search",):
                    v = str(v)
                yield colourise(f"  {k}: {v}", "cyan")
        yield ""
        yield colourise(self.fmt.format(*self.columns.keys()), "white")
        yield colourise(self._get_horizontal_rule(), "white")

    def get_aggregation_header(self, composite: AggregationKey) -> str:
        header = dict(zip(self.aggregate_by, composite))
        values = []
        for name in self.columns:
            if name in self.aggregate_by:
                values.append(header[name])
            else:
                values.append(" ")
        line = self.fmt.format(*values)
        result = re.sub(
            r" ( *)( [^ ]|$)",
            lambda s: " " + len(s.group(1)) * "-" + s.group(2),
            line,
        )
        for name in self.aggregate_by:
            if name not in self.columns:
                result += f" ({name}:{header[name]})"
        return result

    def get_line(self, row: RowDef) -> str:
        values = [
            "-" if val is None else str(val)[: col["width"]]
            for (col, val) in zip(self.columns.values(), row["values"].values())
        ]
        line = self.fmt.format(*values)
        if row.get("colour"):
            line = colourise(line, row["colour"])
        return line

    def get_footer(self) -> Iterable[str]:
        hr = self._get_horizontal_rule()
        yield colourise(hr, "white")
        yield colourise(
            ("{:>" + str(len(hr)) + "}").format(
                "Showing {} of {}".format(
                    min(len(self.rows), self.total_count) or "all",
                    self.total_count,
                )
            ),
            "white",
        )

    def _get_horizontal_rule(self, char="=") -> str:
        """
        A bit of a hack: We get a formatted line for no other reason than to
        determine the width of that line.  Then we use a regex to overwrite that
        line with "=".
        """
        return re.sub(r".", char, self.fmt.format(*["" for c in self.columns]))


class CSVRenderer(Renderer):
    """
    Renderer which outputs comma-separated values, where strings are always
    enclosed in double quotes, and literal double quotes are written as "".
    """

    dialect = "excel"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = StringIO()
        self.writer = csv.writer(self.data, dialect=self.dialect)

    def get_header(self) -> Iterable[str]:
        yield self.get_line({"values": dict((k, k) for k in self.columns.keys())})

    def get_line(self, row: RowDef) -> str:
        self.data.seek(self.data.truncate(0))
        self.writer.writerow(
            [
                "" if v is None else str(v)
                for (k, v) in row["values"].items()
                if k in self.columns
            ]
        )
        return self.data.getvalue()[:-1]


class TabRenderer(Renderer):
    """
    Renderer which outputs tab-separated values, where literal tabs are replaced
    with spaces.
    """

    def get_header(self) -> Iterable[str]:
        yield self.get_line({"values": dict((k, k) for k in self.columns.keys())})

    def get_line(self, row: RowDef) -> str:
        return "\t".join(
            "" if v is None else str(v).replace("\t", "    ")
            for (k, v) in row["values"].items()
            if k in self.columns
        )


class IDAction(argparse.Action):
    def __call__(self, parser, namespace, *args, **kwargs):
        namespace.format = "tab"
        namespace.field = ["id"]
        namespace.no_header = True


def add_argument_group(parser: argparse.ArgumentParser, fields: Iterable[str]):
    """
    Add an argument group to the given parser with all of the parameters
    for controlling the tabular renderers.
    """
    fields = list(fields)
    group = parser.add_argument_group("Output")
    group.add_argument(
        "--field",
        type=str,
        action="append",
        choices=fields,
        default=[],
        help="The field(s) to display. Invoke multiple times for multiple fields.",
    )
    group.add_argument(
        "--format",
        default="pretty",
        choices=renderers.keys(),
        help="Output format. Default: pretty",
    )
    group.add_argument(
        "--no-header",
        action="store_true",
        default=False,
        help="Omit header line(s)",
    )
    group.add_argument(
        "--ids-only",
        nargs=0,
        action=IDAction,
        help="Print only IDs, equivalent to --format=tab --field=id --no-header. "
        "Useful for piping to another command.",
    )
    group.add_argument(
        "--aggregate-by",
        action="append",
        default=[],
        choices=fields,
        help="Aggregate results based on all specified aggregations."
        " Use this option multiple times for more specific aggregations.",
    )
    group.add_argument(
        "--max-per-aggregation",
        type=int,
        help="Maximum number of rows per aggregated bucket.",
    )


renderers = {
    "pretty": PrettyRenderer,
    "csv": CSVRenderer,
    "tab": TabRenderer,
}



# Copyright (c) 2023 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional
from datetime import datetime
import argparse
import importlib
import os
import pkgutil
import sys
import platform

from ..helpers.actions import StoreIfNotEmpty
from ..helpers import xdg
from ..helpers.colours import colourise
from ..version import __version__


class RipeHelpFormatter(argparse.RawTextHelpFormatter):
    def _format_usage(self, *args):
        r = argparse.RawTextHelpFormatter._format_usage(self, *args).capitalize()
        return "\n\n{}\n".format(r)


def _get_command_name(cmd_or_factory):
    return (
        cmd_or_factory.NAME
        if cmd_or_factory.NAME
        else cmd_or_factory.__module__.rsplit(".", 1)[-1]
    )


class Command(object):
    # If NAME isn't defined then the module name is used
    # It exists basically to allow hyphens in the command names
    NAME: Optional[str] = None
    DESCRIPTION = ""  # Define this in the subclass
    _commands = None

    DEPRECATED_ALIASES = {
        "measurement": "measurement-info",
        "measurements": "measurement-search",
        "probe": "probe-info",
        "probes": "probe-search",
        "render": "report",
    }

    def __init__(self, *args, **kwargs):

        self.arguments = None
        self.parser = argparse.ArgumentParser(
            formatter_class=RipeHelpFormatter,
            description="\n\n".join(
                [
                    self.DESCRIPTION + ".",
                    getattr(self, "EXTRA_DESCRIPTION", ""),
                ]
            ),
            prog="ripe-atlas {}".format(self.get_name()),
        )
        self.user_agent = self._get_user_agent()

    @classmethod
    def get_name(cls):
        return _get_command_name(cls)

    @staticmethod
    def _get_packages_for_paths(paths):
        """
        Yield path, package_name for all of the packages found in `paths`.
        """
        for loader, package_name, _ in pkgutil.iter_modules(paths):
            yield loader.path, package_name

    @classmethod
    def _get_user_command_path(cls):
        return os.path.join(xdg.get_config_home(), "commands")

    @classmethod
    def _load_commands(cls):
        """
        Scan for available commands and store a map of command names to module
        paths.
        """
        builtin_path = os.path.dirname(__file__)
        user_command_path = cls._get_user_command_path()

        cls._commands = {}

        paths = [builtin_path, user_command_path]
        for path, package_name in cls._get_packages_for_paths(paths):
            if package_name == "base":
                continue
            if path == builtin_path:
                module = "ripe.atlas.tools.commands.{}".format(package_name)
            else:
                module = package_name
                if user_command_path not in sys.path:
                    sys.path.append(user_command_path)
            cls._commands[package_name] = module

    @classmethod
    def load_command_class(cls, command_name):
        """
        Get the Command or Factory with the given command name.
        """
        if command_name in cls.DEPRECATED_ALIASES:
            alias = command_name
            command_name = cls.DEPRECATED_ALIASES[alias]
            sys.stderr.write(
                colourise(
                    "Warning: {} is a deprecated alias for {}\n\n".format(
                        alias,
                        command_name,
                    ),
                    "yellow",
                )
            )

        if cls._commands is None:
            cls._load_commands()

        try:
            module_name = cls._commands[command_name.replace("-", "_")]
        except KeyError:
            return

        module = importlib.import_module(module_name)

        if hasattr(module, "Factory"):
            cmd = module.Factory
        else:
            cmd = module.Command

        return cmd

    @classmethod
    def get_available_commands(cls):
        """
        Get a list of commands that we can execute.  By default, we have a
        fixed list that we make available in this directory, but the user can
        create her own plugins and store them at
        ~/.config/ripe-atlas-tools/commands/.  If we find any files there, we
        add them to the list here.
        """
        if not cls._commands:
            cls._load_commands()

        return sorted(cls._commands.keys())

    def init_args(self, args=None):
        """
        Initialises all parse arguments and makes them available to the class.
        """

        if args is None:
            args = sys.argv[1:]

        self.arguments = self.parser.parse_args(self._modify_parser_args(args))

    def run(self):
        raise NotImplementedError()

    def add_arguments(self):
        """

        A hook that's executed in the __init__, you can make use of
        `self.parser` here to add arguments to the command:

          self.parser.add_argument(
            "measurement_id",
            type=int,
            help="The measurement id you want to use"
          )

        """
        pass

    def _modify_parser_args(self, args):
        """
        A modifier hook that can be overridden in the child class to allow that
        class to manipulate the arguments before being parsed.  The common
        use-case we're trying to solve here is popping a secondary argument off
        of the list and/or appending `--help` in some circumstances.
        """

        self.add_arguments()

        return args

    def print(self, message):
        """
        Write to stdout regardless of whether it is a tty or not.
        """
        sys.stdout.write(message + "\n")

    def ok(self, message):
        """
        Write a successful message to the tty if present.
        """
        if sys.stdout.isatty():
            sys.stdout.write("\n{}\n\n".format(colourise(message, "green")))

    def not_ok(self, message):
        """
        Writer an unsuccessful message to the tty if present.
        """
        if sys.stdout.isatty():
            sys.stdout.write("\n{}\n\n".format(colourise(message, "red")))

    @staticmethod
    def _get_os_string():
        """
        Return a string identifying the system OS suitable for use in the User-Agent
        header.
        """
        os = platform.system()

        if os == 'Darwin':
            release = platform.mac_ver()[0]
            return f'macOS {release}'
        elif os == 'Windows':
            release = platform.win32_ver()[0]
            return f"Windows {release}"
            pass
        else:
            try:
                # Use a shim for Python < 3.10
                info = xdg.freedesktop_os_release()
            except OSError:
                pass
            else:
                name = info.get("NAME")
                if name:
                    version = info.get("VERSION_ID", "")
                    return f"{name} {version}"
        return platform.platform()

    @classmethod
    def _get_user_agent(cls):
        """
        Allow packagers to change the user-agent to whatever they like by
        placing a file called `user-agent` into the `tools` directory.  If no
        file is found, we go with a sensible platform-specific default + the version.
        """

        try:
            custom = os.path.join(os.path.dirname(__file__), "..", "user-agent")
            with open(custom) as f:
                return f.readline().strip()[:128]
        except IOError:
            pass  # We go with the default

        os_str = cls._get_os_string()

        return f"RIPE Atlas Tools [{os_str}] {__version__}"

    def add_flag(self, parser, name, default, help, no_help=None):
        """
        Convenience method to create a store_true --foo option along with a --no-foo
        store_false counterpart, including proper indication of which is the default.

        This should be used when it's possible for a flag to have a True default,
        either in the defaults or the custom user YAML.
        """
        dest = name.replace("-", "_")

        if no_help is None:
            no_help = f"Do not {help[0].lower()}{help[1:]}"

        parser.set_defaults(**{dest: default})

        default_true = " (default)" if default else ""
        default_false = " (default)" if not default else ""

        parser.add_argument(
            f"--{name}",
            action="store_true",
            help=f"{help}{default_true}",
        )
        parser.add_argument(
            f"--no-{name}",
            action="store_false",
            dest=dest,
            help=f"{no_help}{default_false}",
        )

    def add_primary_argument(self, name, parser):
        """
        Create a positional argument which acts as an alias for --`name`.
        """
        parser.add_argument(
            name,
            help=f"If present, sets the --{name} argument",
            nargs="?",
            action=StoreIfNotEmpty,
        )


class MetaDataMixin(object):
    @staticmethod
    def _prettify_boolean(boolean):

        checkmark = "\u2714"
        x = "\u2718"

        if boolean:
            return colourise(checkmark, "green")
        return colourise(x, "red")

    @staticmethod
    def _prettify_time(dtime):
        if isinstance(dtime, datetime):
            return "{} UTC".format(dtime.isoformat().replace("T", " "))

        return str(dtime)

    @staticmethod
    def _render_line(header, value):
        print("{}  {}".format(colourise("{:25}".format(header), "bold"), value))


class Factory(object):
    NAME = None

    @classmethod
    def build(cls, *args, **kwargs):
        return object()

    @classmethod
    def get_name(cls):
        return _get_command_name(cls)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tzlocal import get_localzone
from datetime import datetime
from .base import Renderer as BaseRenderer
from ..helpers.colours import colourise


class Renderer(BaseRenderer):
    RENDERS = [BaseRenderer.TYPE_NTP]
    TIME_FORMAT = "%a %b %d %H:%M:%S %Z %Y"

    def on_result(self, result):
        created = result.created.astimezone(get_localzone())
        r = self.get_formatted_response(result)
        if not r:
            r = colourise("No results\n", "red")
        return "\n{}\n{}\n\n{}".format(
            colourise("Probe #{}".format(result.probe_id), "bold"),
            colourise(created.strftime(self.TIME_FORMAT), "bold"),
            r,
        )

    @staticmethod
    def get_formatted_response(result):
        leap = result.leap_second_indicator
        stratum = result.stratum
        v = result.version
        mode = result.mode
        # just here for completeness, flake8 doesn't like when it's not used
        # end_time = result.end_time
        poll = result.poll
        precision = result.precision
        refid = result.reference_id
        ref_time = result.reference_time
        root_delay = result.root_delay
        root_disp = result.root_dispersion

        if not leap and not stratum and not v and not mode:
            return

        if mode != "server":
            print("invalid mode: %s" % mode)

        r = "[NTP] %s -> %s (%s)\n" % (
            result.source_address,
            result.destination_name,
            result.destination_address,
        )
        r += "\tversion: %s, stratum: %s/16\n" % (
            colourise(v, "bold"),
            colourise(stratum, "bold"),
        )
        r += "\trefid: %s\n" % colourise(refid, "bold")
        r += "\tleap: %s, poll: %s, precision: %s\n" % (leap, poll, precision)
        r += "\troot_delay: %s, root_disp: %s\n" % (root_delay, root_disp)
        r += "\tref-time: %s\n\n" % datetime.fromtimestamp(ref_time)

        try:
            for idx, pkt in enumerate(result.packets):
                if not pkt.rtt:
                    r += f"\t[{idx}] *\n]"
                    continue
                r += "\t[%s] %s\n" % (idx, str(pkt))
                r += "\t\ttrans: %s -> recv: %s\n" % (
                    pkt.transmitted_time,
                    pkt.received_time,
                )
                r += "\t\torigin: %s -> final: %s\n\n" % (
                    pkt.origin_time,
                    pkt.final_time,
                )
        except Exception as ex:
            print("Got exception when reading packet: %s" % ex)
            print("Raw: %s" % pkt.raw_data)

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):
    """
    This is meant to be a stub example for what an aggregate renderer might look
    like. If you have ideas as to how to make this better, feel free to send
    along a pull request.
    """

    RENDERS = [BaseRenderer.TYPE_PING]

    def __init__(self, **kwargs):
        BaseRenderer.__init__(self, **kwargs)

        self.target = ""
        self.packet_loss = 0
        self.sent_packets = 0
        self.received_packets = 0
        self.rtts = []
        self.rtts_min = []
        self.rtts_max = []
        self.rtt_types_map = {"min": self.rtts_min, "max": self.rtts_max}

    def collect_stats(self, result):
        """
        Calculates, stores and collects all stats we want from the given
        result.
        """
        if not self.target:
            self.target = result.destination_name
        self.sent_packets += result.packets_sent
        self.received_packets += result.packets_received
        self.collect_min_max_rtts("min", result.rtt_min)
        self.collect_min_max_rtts("max", result.rtt_max)

        self.collect_packets_rtt(result.packets)

    def collect_min_max_rtts(self, rtt_type, rtt):
        """
        Stores the given rtt in the corresponding list (min/max) if rtt is set.
        """
        rtt = rtt
        if not rtt:
            rtt = 0

        self.rtt_types_map[rtt_type].append(rtt)

    def collect_packets_rtt(self, packets):
        """
        Collects all the rrts of given packets and stores them
        in our rtts list.
        """
        for packet in packets:
            rtt = packet.rtt
            if not packet.rtt:
                rtt = 0
            self.rtts.append(rtt)

    def calculate_loss(self):
        """Calculates the total loss between received and sent packets."""
        if not self.sent_packets:
            return 0

        return (1 - float(self.received_packets) / self.sent_packets) * 100

    def mean(self):
        """Calculates the mean of the collected rtts"""
        return round(float(sum(self.rtts)) / max(len(self.rtts), 1), 3)

    def median(self):
        """Calculates the median of the collected rtts"""
        sorted_rtts = sorted(self.rtts)
        index = (len(self.rtts) - 1) // 2
        if len(self.rtts) % 2:
            return sorted_rtts[index]
        else:
            return (sorted_rtts[index] + sorted_rtts[index + 1]) / 2.0

    def on_result(self, result):
        packets = result.packets

        if not packets:
            return "No packets found\n"

        self.collect_stats(result)

        # Because the origin value is more reliable as "from" in v4 and as
        # "packet.source_address" in v6.
        origin = result.origin
        if ":" in origin:
            origin = packets[0].source_address

        times = ", ".join([str(_.rtt) + " ms" for _ in packets])

        return (
            f"{result.packet_size} bytes from {result.destination_address} via "
            f"probe #{result.probe_id} ({origin})"
            f": ttl={packets[0].ttl} times={times}\n"
        )

    def header(self, sample):
        resolved_on = (
            "server"
            if sample.destination_address == sample.destination_name
            else "probe"
        )
        return f"PING {sample.destination_name} (resolved on {resolved_on})\n"

    def footer(self):
        if not self.sent_packets:
            return ""
        self.packet_loss = self.calculate_loss()
        return self.render_template(
            "reports/aggregate_ping.txt",
            target=sanitise(self.target),
            sent=self.sent_packets,
            received=self.received_packets,
            packet_loss=self.packet_loss,
            min=min(self.rtts_min),
            median=self.median(),
            mean=self.mean(),
            max=max(self.rtts_max),
        )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from ripe.atlas.sagan import Result
from ripe.atlas.sagan import ResultParseError
from ripe.atlas.cousteau import ProbeRequest
from ripe.atlas.cousteau import Probe as CProbe

from .exceptions import RipeAtlasToolsException
from .cache import cache
from .settings import conf


class FilterFactory(object):
    @staticmethod
    def create(key, value):
        """Create new filter class based on the key"""
        if key == "asn":
            return ASNFilter(value)
        else:
            return Filter(key, value)


class Filter(object):
    """
    Class that represents filter for results. For now supports only attributes
    of probes property of Result property. It could be extended for any property
    of Result easily.
    """

    def __init__(self, key, value):
        self.key = key
        self.value = value

    def filter(self, result):
        """
        Decide if given result should be filtered (False) or remain on the
        pile of results.
        """
        try:
            attr_value = getattr(result.probe, self.key)
        except AttributeError:
            log = (
                "Cousteau's Probe class does not have an attribute " "called: <{}>"
            ).format(self.key)
            raise RipeAtlasToolsException(log)
        if attr_value == self.value:
            return True

        return False


class ASNFilter(Filter):
    """Class thar represents filter by probes that belong to given ASN."""

    def __init__(self, value):
        key = "asn"
        super(ASNFilter, self).__init__(key, value)

    def filter(self, result):
        asn_v4 = getattr(result.probe, "asn_v4")
        asn_v6 = getattr(result.probe, "asn_v6")
        if self.value in (asn_v4, asn_v6):
            return True

        return False


def filter_results(filters, results):
    """docstring for filter"""
    new_results = []
    for result in results:
        for rfilter in filters:
            if rfilter.filter(result):
                new_results.append(result)
                break

    return new_results


class SaganSet(object):
    """
    An iterable of sagan results with attached probe information that allows
    for filtering by the filters module.
    """

    def __init__(self, iterable=None, probes=()):
        self._probes = probes
        self._iterable = iterable

    def __iter__(self):

        sagans = []

        for line in self._iterable:

            # line may be a dictionary (parsed JSON)
            if hasattr(line, "strip"):
                line = line.strip()

            # Break out when there's nothing left
            if not line:
                break

            try:
                sagan = Result.get(
                    line,
                    on_error=Result.ACTION_IGNORE,
                    on_warning=Result.ACTION_IGNORE,
                )
                if not self._probes or sagan.probe_id in self._probes:
                    sagans.append(sagan)
                if len(sagans) > 100:
                    for sagan in self._attach_probes(sagans):
                        yield sagan
                    sagans = []
            except ResultParseError:
                pass  # Probably garbage in the file

        for sagan in self._attach_probes(sagans):
            yield sagan

    def __next__(self):
        return iter(self).next()

    def next(self):
        return self.__next__()

    def _attach_probes(self, sagans):
        probes = dict(
            [
                (p.id, p)
                for p in Probe.get_many(
                    (s.probe_id for s in sagans)
                )
            ]
        )
        for sagan in sagans:
            sagan.probe = probes[sagan.probe_id]
            yield sagan


class Probe(object):
    """
    A crude representation of the data we get from the API via Cousteau
    """

    EXPIRE_TIME = 60 * 60 * 24 * 30

    @classmethod
    def get(cls, pk):
        """
        Given a single id, attempt to fetch a probe object from the cache.  If
        that fails, do an API call to get it.  Don't use this for multiple
        probes unless you know they're all in the cache, or you'll be in for a
        long wait.
        """
        r = cache.get("probe:{}".format(pk))
        if not r:
            kwargs = {"id": pk, "server": conf["api-server"]}
            probe = CProbe(**kwargs)
            cache.set("probe:{}".format(probe.id), probe, cls.EXPIRE_TIME)
            return probe

    @classmethod
    def get_many(cls, ids):
        """
        Given a list of ids, attempt to get probe objects out of the local
        cache.  Probes that cannot be found will be fetched from the API and
        cached for future use.
        """

        r = []

        fetch_ids = []
        for pk in ids:
            probe = cache.get("probe:{}".format(pk))
            if probe:
                r.append(probe)
            else:
                fetch_ids.append(str(pk))

        if fetch_ids:
            kwargs = {"id__in": fetch_ids, "server": conf["api-server"]}
            for probe in [p for p in ProbeRequest(return_objects=True, **kwargs)]:
                cache.set("probe:{}".format(probe.id), probe, cls.EXPIRE_TIME)
                r.append(probe)

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import json

from .base import Command
from ...settings import conf
from ...exceptions import RipeAtlasToolsException


class SpecMeasureCommand(Command):
    DESCRIPTION = "Create a measurement from a JSON spec and wait for the results"

    def clean_target(self):
        return self.arguments.target

    def add_arguments(self):

        Command.add_arguments(self)

        self.parser.add_argument(
            "json_spec",
            help="JSON object containing the RIPE Atlas measurement spec. "
            "Use @some-file.json to load the spec from a given file, or @- to load the "
            "spec from standard input. "
            "Any optional command-line arguments you specify will be merged into this "
            "spec.",
        )
        self.parser.add_argument(
            "--type",
            help="The 'type' of the measurement to be added to the spec JSON",
        )

    def _read_file(self, filename):
        if filename == "-":
            return sys.stdin.read()
        try:
            with open(filename) as f:
                return f.read()
        except FileNotFoundError:
            raise RipeAtlasToolsException(f"No such file: {filename}")

    def _clean_json_spec(self):
        arg = self.arguments.json_spec
        if arg.startswith("@"):
            arg = self._read_file(arg[1:])

        try:
            spec = json.loads(arg)
        except ValueError:
            raise RipeAtlasToolsException("Spec is invalid JSON")

        if not isinstance(spec, dict):
            raise RipeAtlasToolsException("Spec should be a JSON object")
        return spec

    def _get_measurement_kwargs(self):
        spec = self._clean_json_spec()

        if self.arguments.type:
            self._type = spec["type"] = self.arguments.type
        else:
            self._type = spec.get("type")
        if not self._type:
            raise RipeAtlasToolsException('Spec should contain a "type"')
        if self._type not in conf["specification"]["types"]:
            raise RipeAtlasToolsException(f"Unknown measurement type: {self._type}")

        from_args = super()._get_measurement_kwargs()

        # Allow override of description or use a different default that doesn't
        # bother about the target
        if not self.arguments.description:
            del from_args["description"]
            if "description" not in spec:
                spec["description"] = f"{self._type.title()} measurement"

        # Allow override of "af" otherwise use the default
        if not self.arguments.af and "af" in spec:
            del from_args["af"]

        spec.update(from_args)

        return spec



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
import re
import os
import os.path


def get_config_home():
    """ """
    config_home = os.environ.get("XDG_CONFIG_HOME")
    if config_home is None:
        config_home = os.path.expanduser("~/.config")
    return os.path.join(config_home, "ripe-atlas-tools")


if hasattr(platform, "freedesktop_os_release"):
    freedesktop_os_release = platform.freedesktop_os_release
else:
    # Shim for Python versions < 3.10, taken from CPython

    # freedesktop.org os-release standard
    # https://www.freedesktop.org/software/systemd/man/os-release.html

    # NAME=value with optional quotes (' or "). The regular expression is less
    # strict than shell lexer, but that's ok.
    _os_release_line = re.compile(
        "^(?P<name>[a-zA-Z0-9_]+)=(?P<quote>[\"']?)(?P<value>.*)(?P=quote)$"
    )
    # unescape five special characters mentioned in the standard
    _os_release_unescape = re.compile(r"\\([\\\$\"\'`])")
    # /etc takes precedence over /usr/lib
    _os_release_candidates = ("/etc/os-release", "/usr/lib/os-release")
    _os_release_cache = None

    def _parse_os_release(lines):
        # These fields are mandatory fields with well-known defaults
        # in practice all Linux distributions override NAME, ID, and PRETTY_NAME.
        info = {
            "NAME": "Linux",
            "ID": "linux",
            "PRETTY_NAME": "Linux",
        }

        for line in lines:
            mo = _os_release_line.match(line)
            if mo is not None:
                info[mo.group("name")] = _os_release_unescape.sub(
                    r"\1", mo.group("value")
                )

        return info

    def freedesktop_os_release():
        """Return operation system identification from freedesktop.org os-release"""
        global _os_release_cache

        if _os_release_cache is None:
            errno = None
            for candidate in _os_release_candidates:
                try:
                    with open(candidate, encoding="utf-8") as f:
                        _os_release_cache = _parse_os_release(f)
                    break
                except OSError as e:
                    errno = e.errno
            else:
                raise OSError(
                    errno, f"Unable to read files {', '.join(_os_release_candidates)}"
                )

        return _os_release_cache.copy()


def get_os_string():
    os = platform.system()

    if os == 'Darwin':
        release = platform.mac_ver()[0]
        return f'macOS {release}'
    elif os == 'Windows':
        release = platform.win32_ver()[0]
        return f"Windows {release}"
        pass
    else:
        try:
            info = freedesktop_os_release()
        except OSError:
            pass
        else:
            name = info.get("NAME")
            if name:
                version = info.get("VERSION_ID", "")
                return f"{name} {version}"
    return platform.platform()



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ripe.atlas.cousteau import Measurement, AtlasStream
from ripe.atlas.cousteau.exceptions import APIResponseError

from ..exceptions import RipeAtlasToolsException
from ..renderers import Renderer
from ..streaming import StreamWrapper
from ..helpers.validators import ArgumentType
from ..settings import conf
from .base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "stream"

    DESCRIPTION = "Output the results of a public measurement as they become available"
    EXTRA_DESCRIPTION = "Streaming of non-public measurements is not supported."
    URLS = {
        "detail": "/api/v2/measurements/{0}.json",
    }

    def add_arguments(self):
        self.parser.add_argument(
            "measurement_id",
            type=ArgumentType.msm_id_or_name(),
            help="The measurement id or alias you want streamed",
        )
        self.parser.add_argument(
            "--limit",
            type=int,
            help="The maximum number of results you want to stream",
        )
        self.parser.add_argument(
            "--renderer",
            choices=Renderer.get_available(),
            help="The renderer you want to use. If this isn't defined, an "
            "appropriate renderer will be selected.",
        )
        self.parser.add_argument(
            "--timeout",
            type=float,
            help="Stop streaming after this number of seconds",
        )
        self.parser.add_argument(
            "--send-backlog", action="store_true", help="Send backlog of recent results"
        )

        Renderer.add_arguments_for_available_renderers(self.parser)

    def run(self) -> None:
        try:
            measurement = Measurement(
                id=self.arguments.measurement_id,
                user_agent=self.user_agent,
            )
        except APIResponseError as e:
            raise RipeAtlasToolsException(e.args[0])

        self.ok("Connecting to stream...")
        stream = AtlasStream(base_url=conf["stream-base-url"])
        stream.connect()
        stream.subscribe(
            "result",
            msm=self.arguments.measurement_id,
            sendBacklog=self.arguments.send_backlog,
        )
        renderer = Renderer.get_renderer(
            name=self.arguments.renderer, kind=measurement.type.lower()
        )(arguments=self.arguments)
        renderer.render(
            StreamWrapper(
                stream,
                capture_limit=self.arguments.limit,
                timeout=self.arguments.timeout,
            )
        )
        self.ok("Disconnected from stream")



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import pkgutil
import sys

from ..exceptions import RipeAtlasToolsException
from ..helpers import xdg


class Renderer(object):

    TYPE_PING = "ping"
    TYPE_TRACEROUTE = "traceroute"
    TYPE_DNS = "dns"
    TYPE_SSLCERT = "sslcert"
    TYPE_HTTP = "http"
    TYPE_NTP = "ntp"

    RENDERS = ()

    def __init__(self, *args, **kwargs):
        """
        If "arguments" is in kwargs it can be used to gather renderer's
        optional arguments which have been passed via CLI.
        See also add_arguments().
        """
        self.show_header = True
        self.show_footer = True
        if "arguments" in kwargs:
            self.show_header = kwargs["arguments"].show_header
            self.show_footer = kwargs["arguments"].show_footer

    @classmethod
    def get_available(cls):
        """
        Return a list of renderers available to be used.
        """

        paths = [os.path.dirname(__file__)]
        if "HOME" in os.environ:
            path = xdg.get_config_home()
            sys.path.append(path)
            paths += [os.path.join(path, "renderers")]

        names = []

        for _, module_name, _ in pkgutil.iter_modules(paths):
            if module_name == "base":
                continue
            # Check that we can actually use this renderer, otherwise drop it
            try:
                cls.get_renderer_by_name(module_name)
            except Exception:
                continue
            names.append(module_name)

        return names

    @staticmethod
    def add_common_arguments(parser):
        group = parser.add_argument_group(title="Optional arguments for all renderers")
        group.add_argument(
            "--no-header",
            dest="show_header",
            action="store_false",
            help="Don't show a header/title before rendering results",
        )
        group.add_argument(
            "--no-footer",
            dest="show_footer",
            action="store_false",
            help="Don't show a footer/summary after rendering results",
        )

    @staticmethod
    def add_arguments_for_available_renderers(parser):
        Renderer.add_common_arguments(parser)
        for renderer_name in Renderer.get_available():
            renderer_cls = Renderer.get_renderer_by_name(renderer_name)
            renderer_cls.add_arguments(parser)

    @staticmethod
    def render_template(template, **kwargs):
        """
        A crude templating engine.
        """

        template = os.path.join(os.path.dirname(__file__), "templates", template)

        with open(template) as f:
            return str(f.read()).format(**kwargs)

    @classmethod
    def get_renderer(cls, name=None, kind=None):
        """
        Using the name if you've asked for it specifically, or attempting to
        guess the appropriate renderer based on the kind of measurement, this
        will return a Renderer subclass or None if nothing can be found.
        """

        renderer = None

        if name:
            renderer = cls.get_renderer_by_name(name)

        if not renderer and kind:
            renderer = cls.get_renderer_by_kind(kind)

        if kind:
            cls._test_renderer_accepts_kind(renderer, kind)

        return renderer

    @classmethod
    def get_renderer_by_name(cls, name):
        error_message = f'The renderer you selected, "{name}" could not be found.'

        # User-defined, user-supplied
        r = cls.import_renderer("renderers", name)
        if not r:
            r = cls.import_renderer("ripe.atlas.tools.renderers", name)
        if not r:
            raise RipeAtlasToolsException(error_message)

        return r

    @classmethod
    def get_renderer_by_kind(cls, kind):
        error_message = (
            f'A default renderer for "{kind}" measurements could not be found.'
        )

        r = cls.import_renderer("ripe.atlas.tools.renderers", kind)
        if not r:
            raise RipeAtlasToolsException(error_message)

        return r

    @staticmethod
    def import_renderer(package, name):
        """
        Return the Renderer class from package.name, or None if either package
        or package.name don't exist.
        """
        full_name = f"{package}.{name}"
        try:
            spec = importlib.util.find_spec(full_name)
        except ModuleNotFoundError:
            return
        if not spec:
            return
        return getattr(importlib.import_module(full_name), "Renderer")

    @staticmethod
    def add_arguments(parser):
        """
        Add renderer's optional arguments here.

        Suggested format:

        group = parser.add_argument_group(
            title="Optional arguments for XXX renderer"
        )
        group.add_argument(
            ...
        )
        """
        pass

    def render(self, results, sample=None):
        """
        Render the given iterable of RIPE Atlas JSON results.
        """
        # Put aggregated and unaggregated results in the same format
        normalized = dict(results) if isinstance(results, dict) else {"": results}

        header_shown = False
        last_key = None

        for key, results in normalized.items():
            for sagan in results:
                # Possibly show render header
                if self.show_header and not header_shown:
                    print(self.header(sagan), end="")
                    header_shown = True

                if key:
                    indent = " "
                    if key != last_key:
                        # Show aggregation group header
                        print("\n" + key)
                        last_key = key
                else:
                    indent = ""

                line = Result(self.on_result(sagan), sagan.probe_id)

                print(indent + line, end="")

        if self.show_footer:
            print(self.footer(), end="")

    def header(self, sample):
        """
        Override this to add a header.

        `sample` is a single parsed result from the result set (probably the
        first one). It can be used to infer metadata about the measurement
        without having to do an extra API call.
        """
        return ""

    def footer(self):
        """
        Override this to add a footer.

        To provide a summary here, statistics should be gathered in the
        `on_result` callback, ideally without storing all results in memory.
        """
        return ""

    @staticmethod
    def _test_renderer_accepts_kind(renderer, kind):
        if kind not in renderer.RENDERS:
            raise RipeAtlasToolsException(
                "The renderer selected does not appear to support measurements "
                'of type "{}"'.format(kind)
            )

    def on_result(self, result):
        """
        This must be defined in the subclass, and must return a string, even if
        that string is "".
        """
        raise NotImplementedError()


class Result(str):
    """
    A string-like object that we can use to render results, but that contains
    enough information to be used by the aggregators if need be.
    """

    def __new__(cls, value, probe_id):
        obj = str.__new__(cls, value)
        obj.probe_id = probe_id
        return obj



# Copyright (c) 2023 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterator, Optional

from ripe.atlas.cousteau import AtlasStream
from ripe.atlas.sagan import Result


class StreamWrapper:
    """
    Iterable wrapper for AtlasStream that yields sagan Results up to a
    specified capture limit and/or timeout
    """

    def __init__(
        self,
        stream: AtlasStream,
        capture_limit: Optional[int] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self.stream = stream
        self.capture_limit = capture_limit
        self.timeout = timeout
        self.num_received = 0

    def __iter__(self) -> Iterator[Result]:
        for event_name, payload in self.stream.iter(seconds=self.timeout):
            if event_name == "atlas_result":
                parsed = Result.get(
                    payload,
                    on_error=Result.ACTION_IGNORE,
                    on_malformation=Result.ACTION_IGNORE,
                )
                yield parsed
                self.num_received += 1
                if self.num_received == self.capture_limit:
                    break



# Copyright (c) 2023 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
from typing import Dict

from ripe.atlas.cousteau import MeasurementRequest

from ..helpers import tabular
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType
from ..settings import conf
from .base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "measurement-search"
    LIMITS = (1, 10000)
    MAX_PAGE_SIZE = 500  # Request chunks of this size or smaller

    STATUS_SPECIFIED = 0
    STATUS_SCHEDULED = 1
    STATUS_ONGOING = 2
    STATUS_STOPPED = 4
    STATUS_FORCED_STOP = 5
    STATUS_NO_SUITABLE_PROBES = 6
    STATUS_FAILED = 7
    STATUSES = {
        "scheduled": (
            STATUS_SPECIFIED,
            STATUS_SCHEDULED,
        ),
        "ongoing": (STATUS_ONGOING,),
        "stopped": (
            STATUS_STOPPED,
            STATUS_FORCED_STOP,
            STATUS_NO_SUITABLE_PROBES,
            STATUS_FAILED,
        ),
    }

    COLUMNS: Dict[str, tabular.ColumnDef] = {
        "id": {"align": ">", "width": 7},
        "type": {"align": "^", "width": 10},
        "description": {"align": "^", "width": 42},
        "status": {"align": "^", "width": 18},
        "target": {"align": ">", "width": 45},
        "url": {"align": "^", "width": 45},
    }

    DESCRIPTION = (
        "Fetch and print measurements fulfilling specified criteria based "
        "on given filters"
    )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "--search",
            type=str,
            help="A search string.  This could match the target or description.",
        )
        self.parser.add_argument(
            "--status",
            type=str,
            choices=self.STATUSES.keys(),
            help="The measurement status.",
        )
        self.parser.add_argument(
            "--af", type=int, choices=(4, 6), help="The address family."
        )
        self.parser.add_argument(
            "--type",
            type=str,
            choices=("ping", "traceroute", "dns", "sslcert", "ntp", "http"),
            help="The measurement type.",
        )

        timing = self.parser.add_argument_group("Timing")
        for position in ("started", "stopped"):
            for chrono in ("before", "after"):
                timing.add_argument(
                    "--{}-{}".format(position, chrono),
                    type=ArgumentType.datetime,
                    help="Filter for measurements that {} {} a specific date. "
                    "The format required is YYYY-MM-DDTHH:MM:SS".format(
                        position, chrono
                    ),
                )

        self.parser.add_argument(
            "--limit",
            type=ArgumentType.integer_range(self.LIMITS[0], self.LIMITS[1]),
            default=50,
            help="The number of measurements to return.  The number must be "
            "between {} and {}".format(self.LIMITS[0], self.LIMITS[1]),
        )
        tabular.add_argument_group(self.parser, self.COLUMNS.keys())

    def run(self) -> None:
        if not self.arguments.field:
            self.arguments.field = ["id", "type", "description", "status"]

        request_fields = list(
            set(self.arguments.field) | set(self.arguments.aggregate_by)
        )
        if "status" not in request_fields:
            request_fields.append("status")

        filters = self._get_filters()
        measurements = MeasurementRequest(
            server=conf["api-server"],
            return_objects=True,
            user_agent=self.user_agent,
            fields=",".join(request_fields),
            page_size=min(self.MAX_PAGE_SIZE, self.arguments.limit),
            **filters,
        )
        truncated_measurements = itertools.islice(measurements, self.arguments.limit)

        renderer = tabular.renderers[self.arguments.format]

        rows = [self._get_row(m) for m in truncated_measurements]

        for line in renderer(
            rows=rows,
            total_count=measurements.total_count,
            columns=dict((c, self.COLUMNS[c]) for c in self.arguments.field),
            filters=filters,
            arguments=self.arguments,
        ):
            print(line)

    def _get_row(self, measurement) -> tabular.RowDef:
        r = {}

        for field in self.arguments.field + self.arguments.aggregate_by:
            if field == "url":
                r["url"] = f"https://atlas.ripe.net/measurements/{measurement.id}/"
            elif field == "type":
                r["type"] = measurement.type.lower()
            elif field == "target":
                r["target"] = sanitise(
                    measurement.target or measurement.target_ip or "-"
                )
            elif field == "description":
                r["description"] = sanitise(measurement.description) or ""
            else:
                r[field] = sanitise(getattr(measurement, field))

        return {
            "values": r,
            "colour": self._get_colour_from_status(measurement.status_id),
        }

    def _get_filters(self) -> Dict[str, str]:

        r = {}

        if self.arguments.search:
            r["search"] = self.arguments.search
        if self.arguments.status:
            r["status__in"] = self.STATUSES[self.arguments.status]
        if self.arguments.af:
            r["af"] = self.arguments.af
        if self.arguments.type:
            r["type"] = self.arguments.type
        if self.arguments.started_before:
            r["start_time__lt"] = self.arguments.started_before
        if self.arguments.started_after:
            r["start_time__gt"] = self.arguments.started_after
        if self.arguments.stopped_before:
            r["stop_time__lt"] = self.arguments.stopped_before
        if self.arguments.stopped_after:
            r["stop_time__gt"] = self.arguments.stopped_after

        return r

    def _get_colour_from_status(self, status: int) -> str:
        if status in self.STATUSES["ongoing"]:
            return "green"
        if status == self.STATUS_STOPPED:
            return "yellow"
        if status in self.STATUSES["stopped"]:
            return "red"
        if status in self.STATUSES["scheduled"]:
            return "blue"
        return "white"



from argparse import Action


class StoreIfNotEmpty(Action):
    """
    Like 'store' but don't overwrite an existing/conflicting option if the given
    value is empty.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            setattr(namespace, self.dest, values)



# Copyright (c) 2023 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import sys
from typing import Any, Dict, List, Mapping, Tuple

import requests
from ripe.atlas.cousteau import ProbeRequest

from ..exceptions import RipeAtlasToolsException
from ..helpers import tabular
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType
from ..settings import conf
from .base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "probe-search"
    MAX_PAGE_SIZE = 500  # Request chunks of this size or smaller

    DESCRIPTION = (
        "Fetch and print probes fulfilling specified criteria based on " "given filters"
    )

    COLUMNS: Dict[str, tabular.ColumnDef] = {
        "id": {"align": ">", "width": 7},
        "asn_v4": {"align": ">", "width": 6},
        "asn_v6": {"align": ">", "width": 6},
        "country": {"align": "^", "width": 7},
        "status": {"align": "^", "width": 15},
        "prefix_v4": {"align": ">", "width": 18},
        "prefix_v6": {"align": ">", "width": 18},
        "coordinates": {"align": "^", "width": 19},
        "is_public": {"align": "^", "width": 9},
        "description": {"align": "^", "width": 30},
        "address_v4": {"align": ">", "width": 15},
        "address_v6": {"align": ">", "width": 39},
        "is_anchor": {"align": "^", "width": 9},
    }

    def add_arguments(self) -> None:
        """Adds all commands line arguments for this command."""
        asn = self.parser.add_argument_group("ASN")
        asn.add_argument(
            "--asn",
            type=int,
            help="Probes in IPv4 or IPv6 prefixes announced by this ASN",
        )
        asn.add_argument(
            "--asnv4", type=int, help="Probes in IPv4 prexfixes announced by this ASN"
        )
        asn.add_argument(
            "--asnv6", type=int, help="Probes in IPv6 prefixes announced by this ASN"
        )

        prefix = self.parser.add_argument_group("Prefix")
        prefix.add_argument(
            "--prefix",
            type=str,
            help="Probes with addresses in this IPv4 or IPv6 CIDR prefix",
        )
        prefix.add_argument(
            "--prefixv4", type=str, help="Probes with addresses in this IPv4 prefix"
        )
        prefix.add_argument(
            "--prefixv6", type=str, help="Probes with addresses in this IPv6 prefix"
        )

        area = self.parser.add_argument_group("Area")
        geo_location = area.add_mutually_exclusive_group()
        geo_location.add_argument(
            "--location",
            type=str,
            help="The location of probes as a string i.e. 'Amsterdam'",
        )
        geo_location.add_argument(
            "--center",
            type=str,
            help="Location as <lat>,<lon>-string, i.e. '48.45,9.16'. "
            "Note: use --center=-5,10 (no space) to allow for negative latitudes",
        )
        geo_location.add_argument(
            "--country", type=str, help="The country code of probes."
        )
        area.add_argument(
            "--radius",
            type=int,
            default=15,
            help="Radius in km from specified center/point. Default: 15",
        )

        self.parser.add_argument(
            "--tag",
            type=ArgumentType.tag,
            action="append",
            metavar="TAG",
            help="Include only probes that are marked with these tags. "
            "Use --tag multiple times to filter on the basis of more "
            "than one tag. "
            "Example: --tag system-ipv6-works --tag system-ipv4-works",
            dest="tags",
        )

        self.parser.add_argument(
            "--limit",
            type=int,
            default=25,
            help="Return at most this number of probes. Default: 25",
        )
        self.parser.add_argument(
            "--all",
            action="store_true",
            help="Fetch all probes; takes a long time!",
        )
        self.parser.add_argument(
            "--status",
            type=int,
            choices=[0, 1, 2, 3],
            help=(
                "Probe's connection status [0 - Never Connected, "
                "1 - Connected, 2 - Disconnected, 3 - Abandoned]"
            ),
        )
        self.parser.add_argument(
            "--auth",
            type=str,
            default=conf["authorisation"]["google_geocoding"],
            help=(
                "Google Geocoding API key to be " "used to perform --location search."
            ),
        )
        tabular.add_argument_group(self.parser, self.COLUMNS.keys())

    def run(self) -> None:
        if not self.arguments.field:
            self.arguments.field = [
                "id",
                "asn_v4",
                "asn_v6",
                "country",
                "status",
            ]

        if self.arguments.all:
            self.arguments.limit = sys.maxsize

        filters = self._get_filters()
        request_fields = self._get_request_fields()

        probes = ProbeRequest(
            server=conf["api-server"],
            return_objects=True,
            user_agent=self.user_agent,
            fields=",".join(request_fields),
            page_size=min(self.MAX_PAGE_SIZE, self.arguments.limit),
            **filters
        )
        truncated_probes = itertools.islice(probes, self.arguments.limit)

        renderer = tabular.renderers[self.arguments.format]

        rows = [self._get_row(m) for m in truncated_probes]

        for line in renderer(
            rows=rows,
            total_count=probes.total_count,
            columns=dict((c, self.COLUMNS[c]) for c in self.arguments.field),
            filters=filters,
            arguments=self.arguments,
        ):
            print(line)

    def _get_filters(self) -> Dict[str, str]:
        """
        Get the request filters for sending to the API.
        """
        if self.arguments.all:
            return {}

        args: Dict[str, Any] = {}

        if any([self.arguments.asn, self.arguments.asnv4, self.arguments.asnv6]):
            args.update(self._clean_asn())

        if any(
            [
                self.arguments.prefix,
                self.arguments.prefixv4,
                self.arguments.prefixv6,
            ]
        ):
            args.update(self._clean_prefix())

        if self.arguments.location:
            args.update(self._clean_location())

        if self.arguments.center:
            args.update(self._clean_center())

        if self.arguments.country:
            args.update(self._clean_country_code())

        if self.arguments.status is not None:
            args.update({"status": self.arguments.status})

        if self.arguments.tags:
            args.update({"tags": ",".join(self.arguments.tags)})

        return args

    def _clean_asn(self) -> Mapping[str, int]:
        """Make sure ASN arguments don't conflict and make sense."""
        asn = self.arguments.asn
        asnv4 = self.arguments.asnv4
        asnv6 = self.arguments.asnv6

        if asn and (asnv4 or asnv6):
            exc_log = (
                "Specifying argument --asn together with --asnv4/--asnv6 "
                "doesn't make sense"
            )
            raise RipeAtlasToolsException(exc_log)
        if asn:
            return {"asn": asn}

        asn_args = {}
        if asnv4:
            asn_args["asn_v4"] = asnv4

        if asnv6:
            asn_args["asn_v6"] = asnv6

        return asn_args

    def _clean_prefix(self) -> Dict[str, str]:
        """Make sure ASN arguments don't conflict and make sense."""
        prefix = self.arguments.prefix
        prefixv4 = self.arguments.prefixv4
        prefixv6 = self.arguments.prefixv6

        if prefix and (prefixv4 or prefixv6):
            exc_log = (
                "Specifying argument --prefix together with "
                "--prefixv4/--prefixv6 doesn't make sense"
            )
            raise RipeAtlasToolsException(exc_log)
        if prefix:
            return {"prefix": prefix}

        prefix_args = {}
        if prefixv4:
            prefix_args["prefix_v4"] = prefixv4

        if prefixv6:
            prefix_args["prefix_v6"] = prefixv6

        return prefix_args

    def _clean_location(self) -> Dict[str, Any]:
        """Make sure location argument are sane."""
        if not self.arguments.auth:
            raise RipeAtlasToolsException(
                "--location requires a Google Geocoding API  key specified with "
                "--auth or configure command (authorisation.google_geocoding)"
            )
        lat, lng = self.location2degrees()
        if self.arguments.radius:
            location_args = {
                "radius": "{0},{1}:{2}".format(lat, lng, self.arguments.radius)
            }
        else:
            location_args = {"latitude": lat, "longitude": lng}

        return location_args

    def location2degrees(self) -> Tuple[str, str]:
        """Fetches degrees based on the given location."""
        error_log = (
            "The following error occured while trying to fetch lat/lon "
            "for location <{}>:\n\n{}"
        )
        google_api_url = "https://maps.googleapis.com/maps/api/geocode/json"
        try:
            result = requests.get(
                google_api_url,
                params={
                    "key": self.arguments.auth,
                    "address": self.arguments.location,
                },
            )
        except (
            requests.ConnectionError,
            requests.HTTPError,
            requests.Timeout,
        ) as e:
            error_log = error_log.format(self.arguments.location, e)
            raise RipeAtlasToolsException(error_log)

        data = result.json()

        if "error_message" in data:
            error = error_log.format(self.arguments.location, data["error_message"])
            raise RipeAtlasToolsException(error)

        try:
            lat = data["results"][0]["geometry"]["location"]["lat"]
            lng = data["results"][0]["geometry"]["location"]["lng"]
        except (KeyError, IndexError) as e:
            error = error_log.format(self.arguments.location, e)
            raise RipeAtlasToolsException(error)

        return str(lat), str(lng)

    def _clean_center(self) -> Dict[str, Any]:
        """Make sure center argument are sane."""
        try:
            lat, lng = self.arguments.center.split(",")
        except ValueError:
            raise RipeAtlasToolsException(
                "Point argument should be in <lat,lng> format."
            )

        if self.arguments.radius:
            center_args = {
                "radius": "{0},{1}:{2}".format(lat, lng, self.arguments.radius)
            }
        else:
            center_args = {"latitude": lat, "longitude": lng}

        return center_args

    def _clean_country_code(self) -> Dict[str, str]:
        """Make sure country_code argument are sane."""
        return {"country_code": self.arguments.country}

    def _get_row(self, probe) -> tabular.RowDef:
        r = {}

        for field in self.arguments.field + self.arguments.aggregate_by:
            if field == "country":
                r[field] = (probe.country_code or "").lower()
            elif field in ("asn_v4", "asn_v6"):
                r[field] = getattr(probe, field) or None
            elif field == "description":
                description = sanitise(probe.description) or None
                r[field] = description
            elif field == "coordinates":
                if probe.geometry and probe.geometry["coordinates"]:
                    lng, lat = probe.geometry["coordinates"]
                    r[field] = "{:7.4f} {:8.4f}".format(lat, lng)
                else:
                    r[field] = None
            elif field in ("is_public", "is_anchor"):
                if getattr(probe, field):
                    r[field] = "\u2714"  # Check mark
                else:
                    r[field] = "\u2718"  # X
            else:
                r[field] = sanitise(getattr(probe, field))

        return {"values": r, "colour": self._get_colour_from_status(probe.status)}

    @staticmethod
    def _get_colour_from_status(status: str) -> str:
        if status == "Connected":
            return "green"
        if status == "Disconnected":
            return "yellow"
        if status == "Abandoned":
            return "red"
        return "white"

    def _get_request_fields(self) -> List[str]:
        request_fields = list(self.arguments.field)
        for field in self.arguments.aggregate_by:
            if field not in request_fields:
                request_fields.append(field)
        if "country" in request_fields:
            request_fields.remove("country")
            request_fields.append("country_code")
        if "status" not in request_fields:
            request_fields.append("status")
        if "coordinates" in request_fields:
            request_fields.remove("coordinates")
            request_fields.append("geometry")
        return request_fields
